//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//                                                                    
//     Jacobi method for TEST PROBLEM 1                               
//                            parallelized with OpenMP                
//                                                                    
//--------------------------------------------------------------------
//                                                                    
//     Sample Code for Lecture in NR/HEP2 Spring School               
//                                                                    
//                                     Coded by Hirotada Okawa        
//                                                                    
//====================================================================
//   How to compile and use this program in terminal(bash)
//====================================================================
//   $ g++ -O2 -ffast-math -fopenmp -o j_ompc jacobi_openMP.C
//   $ export OMP_NUM_THREADS=2
//   $ ./j_ompc
//--------------------------------------------------------------------
//   OMP_NUM_THREADS : Change the number of cores you want to use.
//--------------------------------------------------------------------
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

//-------------------------------------------------------------
//     Include Basic libraries                                
//-------------------------------------------------------------
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>

using namespace std;

int main(int argc, char* argv[]){

  //-----------------------------------------------------------
  //     Grid points
  //-----------------------------------------------------------
  const int jlip=1, juip=100, jlbp=jlip-1, jubp=juip+1;
  //-----------------------------------------------------------
  //     For Generalization of C++ array
  //-----------------------------------------------------------
  const int jli=jlip-jlbp, jui=juip-jlbp, jlb=0, jub=jubp-jlbp;
  const int Nj=jub-jlb+1;

  //-----------------------------------------------------------
  //     Maximum/Minimum Coordinates (physical)
  //-----------------------------------------------------------
  const double xlower=0.;
  const double xupper=1.;
  const double dx=(xupper-xlower)/static_cast<double>(jub-jli);
  const double dxi=1./dx;

  //-----------------------------------------------------------
  //     Array for Coordinates (physical)
  //-----------------------------------------------------------
  double *x;

  //-----------------------------------------------------------
  //     Variables to solve
  //-----------------------------------------------------------
  double *h, *hprev;

  //-----------------------------------------------------------
  //     Source term for Poisson equation
  //-----------------------------------------------------------
  double *src;

  //-----------------------------------------------------------
  //     Definition of parameters
  //-----------------------------------------------------------
  const int stepmax=1e8;
  const double errormax=1.e-10;
  const double fpar=1.;
  const double omega=1.;

  //-----------------------------------------------------------
  //     Definition of temporary variables to use
  //-----------------------------------------------------------
  int j, step=0;
  double xx;
  double errortmp=0.;
  double vtmp=0.;

  //-----------------------------------------------------------
  //     Output File
  //-----------------------------------------------------------
  ofstream ofs;
  ofs.open("h_oc.dat");

  //-----------------------------------------------------------
  //     Initialization
  //-----------------------------------------------------------
  x     = new double[Nj];
  h     = new double[Nj];
  hprev = new double[Nj];
  src   = new double[Nj];

  //-----------------------------------------------------------
  //     OpenMP threads folk
  //-----------------------------------------------------------
#pragma omp parallel default(shared) private(j,xx)
  {

#pragma omp for
  for(j=jlb;j<=jub;j++){
    x[j] = xlower +(static_cast<double>(j)-0.5)*dx;
    h[j] = 1.;
    hprev[j] = h[j];
  }

  //***********************************************************
  //     Main Loop
  //***********************************************************
  do{

    //---------------------------------------------------------
    //     Preserve data of previous step
    //---------------------------------------------------------
#pragma omp for
    for(j=jlb;j<=jub;j++){
      hprev[j] = h[j];
    }

    //---------------------------------------------------------
    //     Jacobi Method
    //---------------------------------------------------------
#pragma omp for
    for(j=jli;j<=jui;j++){

      //=======================================================
      //     Definition of Source term
      //=======================================================
      xx = x[j];
      src[j] = xx*xx*12.;

      h[j] = 0.5*( hprev[j+1] +hprev[j-1] -dx*dx*src[j] ); 
    }

    //=========================================================
    //     Impose Boundary Condition
    //=========================================================
#pragma omp single
    {
      h[jub]=pow(x[jub],4);    // Dirichlet Boundary Condition
      h[jlb]=h[jli];           // Neumann Boundary Condition
    }

    //---------------------------------------------------------
    //     Check if values converge
    //---------------------------------------------------------
    errortmp=0.;
    vtmp=0.;
#pragma omp barrier
#pragma omp for reduction(+:vtmp,errortmp)
    for(j=jli;j<=jui;j++){
      errortmp += (h[j]-hprev[j])*(h[j]-hprev[j])*dx*dx;
      vtmp += dx*dx;
    }

#pragma omp single
    {
      errortmp = sqrt(errortmp/vtmp);
    }
    if( (errortmp<=errormax) && (step > 1) ) break;

    //---------------------------------------------------------
    //     Next Guress(Mix Previous value) just in case
    //---------------------------------------------------------
#pragma omp for
    for(j=jli;j<=jui;j++){
      h[j] = fpar*h[j] +(1.-fpar)*hprev[j];
    }

#pragma omp single
    {
      cout << "Step=" << step << " Error=" << errortmp << endl;
      step++;
    }
  }while(step<stepmax);

  }

  //-----------------------------------------------------------
  //     Print Data
  //-----------------------------------------------------------
  ofs.setf(ios_base::scientific);
  ofs.precision(8);
  for(j=jlb;j<=jub;j++){
    ofs << setw(16) << x[j]
	<< setw(16) << h[j]
	<< setw(16) << hprev[j]
	<< setw(16) << src[j]
	<< endl;
  }

  //-----------------------------------------------------------
  //     End of Program
  //-----------------------------------------------------------
  cout << "End of Run " << errortmp << endl;
  ofs.close();

  delete[] x;
  delete[] h;
  delete[] hprev;
  delete[] src;

  return 0;
}
















